<?php
/**
 * localfile - class for manipulation of local file with locking
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Haruki Setoyama <haruki@planewave.org>
 * @copyright Copyright &copy; 2004, Haruki SETOYAMA <haruki@planewave.org>
 * @license http://www.opensource.org/licenses/lgpl-license.php LGPL
 * @version $Id: localfile.php,v 1.1 2004/06/17 09:00:27 haruki Exp $
 * @package localfile
 */
/**
 * Requires PEAR
 */
require_once 'PEAR.php';
/**
 * error code
 */
define('LOCALFILE_ERROR', 1);

/**
 * localfile - class for manipulation of local file with locking
 *
 * @access public
 * @package localfile
 */
class LocalFile extends PEAR {

    /**
     * flag for use of flock().
     */
    var $flock = true;

    /**
     * file pointer for lock file
     * @access private
     */
    var $_lfp;

    /**
     * file pointer for lock file
     * @access private
     */
    var $_file;

    /**
     * if file_get_contents() function exists or not
     * @access private
     */
    var $_file_get_contents;

    /**
     * constructor
     * @param string $file
     */
    function LocalFile($file)
    {
        $this->_file = $file;
        $this->_file_get_contents = function_exists('file_get_contents');
    }

    /**
     * read content of the file
     * @return string
     */
    function read($new = false)
    {
        $lock = $this->lock(LOCK_SH);
        if ($lock !== true) { return $err; }

        if ($new || ! file_exists($this->_file)) {
            $ret = touch($this->_file);
            if ($ret === false) {
                $this->unlock();
                return $this->raiseError(
                    "Failed to create a file '".$this->_file."'. ",
                    LOCALFILE_ERROR);
            }
            return '';
        }

        if ($this->_file_get_contents) {
            $content = file_get_contents($this->_file);
        } else {
            $fp = fopen($this->_file, 'rb');
            if ($fp === false) {
                $this->unlock();
                return $this->raiseError(
                    "Failed to open a file '".$this->_file."' for reading. ",
                    LOCALFILE_ERROR);
            }
            $content = fread($fp, filesize($this->_file));
        }
        $this->unlock();
        if ($content === false) {
            return $this->raiseError(
                "Failed to get content from a file '".$this->_file."'. ",
                LOCALFILE_ERROR);
        }
        return $content;
    }

    /**
     * write $content to file
     * @param string
     */
    function write($content, $mode =null)
    {
        // file lock.
        $lock = $this->lock();
        if ($lock !== true) { return $err; }

        $tempf = tempnam(dirname($this->_file), 'tmp.');
        $fp_w = fopen($tempf, 'wb');
        if (! $fp_w) {
            $this->unlock();
            return $this->raiseError(
                "Failed to open a temporary file '$tempf' for writting. ",
                LOCALFILE_ERROR);
        }
        $err = fwrite($fp_w, $content);
        fclose($fp_w);
        if (! $err) {
            unlink($tempf);
            $this->unlock();
            return $this->raiseError(
                "Failed to write a content to a temporary file '$tempf'. ",
                LOCALFILE_ERROR);
        }

        if (file_exists($this->_file)) {
            $err = unlink($this->_file);
            if (! $err) {
                unlink($tempf);
                $this->unlock();
                return $this->raiseError(
                    "Failed to unlink a file '".$this->_file."'. ",
                    LOCALFILE_ERROR);
            }
        }
        $err = rename($tempf, $this->_file);
        if ($err === false) {
            unlink($tempf);
            $this->unlock();
            return $this->raiseError(
                "Failed to rename the temporary file to '".$this->_file."'. ",
                LOCALFILE_ERROR);
        }
        if (isset($mode)) {
            chmod($this->_file, $mode);
        }
        $this->unlock();
        return true;
    }

    /**
     * lock file
     */
    function lock($type = LOCK_EX)
    {
        if ($this->flock != false) {
            if (isset($this->_lfp)) {
                return true;
            }
            $lock = $this->_file.'.lock';
            $this->_lfp = fopen($lock, 'wb');
            if (! $this->_lfp) {
                return $this->raiseError(
                    "Failed to open lockfile '$lock' for writing. ",
                    LOCALFILE_ERROR);
            }
            if (! flock($this->_lfp, LOCK_EX)) {
                fclose($this->_lfp);
                unset($this->_lfp);
                return $this->raiseError(
                    "Failed to flock '$lock'. ",
                    LOCALFILE_ERROR);
            }
        }
        return true;
    }

    /**
     * unlock  file
     */
    function unlock()
    {
        if ($this->flock != false && isset($this->_lfp)) {
            $lock = $this->_file.'.lock';
            fclose($this->_lfp);
            unlink($lock);
        }
        return true;
    }
}

?>