<?php
/**
 * zogan - PHP template engine for XML
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Haruki Setoyama <haruki@planewave.org>
 * @copyright Copyright &copy; 2003-2004, Haruki SETOYAMA <haruki@planewave.org>
 * @license http://www.opensource.org/licenses/lgpl-license.php LGPL
 * @version 0.1- $Id: zogan.elements.php,v 1.2 2004/06/17 12:55:56 haruki Exp $
 * @link http://phpoot.sourceforge.jp/zogan/
 * @package zogan
 * @subpackage _builder
 */
/**
* Base class of template builder for each contents of tag
* This class and its sub classes are accessed only by zogan_parser class.
* @see zogan_parser
* @package zogan
* @subpackage _builder
* @abstract
*/
class zogan_content {
    /**
     * indent string at the beginning of the content
     * @var string
     */
    var $indent = '';

    /**
     * line feed at the end of the content
     * @var string
     */
    var $linefeed = '';

    /**
     * flag about that this content is removed or not
     * @var bool
     */
    var $removed = false;

    /**
     * constructor
     */
    function zogan_content($data = '') {}

    /**
     * build PHP script
     * @return string
     */
    function build()
    {
        return '';
    }

    /**
     * returns if this content needs indent and linefeed
     * @return string
     */
    function need_indent()
    {
        return false;
    }

    /**
     * returns if this content consites of spaces only or is empty
     * @return string
     */
    function is_space()
    {
        return false;
    }

    /**
     * returns if this class can have alternative content(s)
     * @return array|false
     */
    function is_alternative()
    {
        return false;
    }

    /**
     * returns if this content represent format to display
     * @return array|false
     */
    function formatable()
    {
        return false;
    }

    /**
     * double quote for html attribute
     * @access private
     * @return string
     */
    function _double_quote_xml($str)
    {
        return '"' . str_replace(array('"', '<', '>'), array('&quot', '&lt;', '&gt;'), $str) . '"';
    }

    /**
     * single quote for php
     * @access private
     * @return string
     */
    function _single_quote($str)
    {
        return '\'' . str_replace('\'', '\\\'', $str) . '\'';
    }

    function _double_quote($str)
    {
        return '"' . str_replace("\n", '\\n', $str) . '"';
    }
}

/**
* null object
*/
class zogan_content_null extends zogan_content {

    function zogan_content_null($data = '') {}

    function is_space()
    {
        return true;
    }
}

/**
* for charactor data
*/
class zogan_content_cdata extends zogan_content {

    var $_content = '';

    function zogan_content_cdata($data)
    {
        $this->_content = $data;
    }

    function build()
    {
        return $this->indent . $this->_content . $this->linefeed;
    }

    function is_space()
    {
        return (! preg_match('/[^\040\011]/', $this->_content));
    }

    function formatable()
    {
        $data = trim($this->_content);
        if (preg_match('/^\{([a-z_]+)(?:\s+"([^"]*)")?\}$/D', $data, $match)
        && file_exists(ZOGAN_FORMAT . $match[1]. '.php')) {
            return array($match[1], $match[2]);
        }
        return false;
    }
}

/**
* default content returns data without modified
*/
class zogan_content_default extends zogan_content {

    var $_content = '';

    function zogan_content_default($data)
    {
        $this->_content = $data;
    }

    function build()
    {
        return $this->indent .
            "<?php echo ".$this->_single_quote($this->_content)." ?>\n"
            . $this->linefeed;
    }
}

/**
* for linefeed
*/
class zogan_content_linefeed extends zogan_content_cdata {
    function zogan_content_linefeed($data)
    {
        parent::zogan_content_cdata($data);
    }

    function is_space()
    {
        return true;
    }
}


/**
* content for escape
*/
class zogan_content_escape extends zogan_content {

    var $_content;

    function zogan_content_escape($data)
    {
        $this->_content = $data;
    }

    function build()
    {
        return $this->indent . '<!' . $this->_content . '>' . $this->linefeed;
    }
}

/**
* content for pi
*/
class zogan_content_pi extends zogan_content {

    var $_content;
    var $_target;

    function zogan_content_pi($target, $data)
    {
        $this->_content = $data;
        $this->_target = $target;
    }

    function build()
    {
        $data = '<?' . $this->_target . ' ' . $this->_content . '?>';
        if ($this->_target == 'php') {
            return $this->indent . $data . $this->linefeed;
        } else {
            return $this->indent . "<?php echo ".$this->_single_quote($data)." ?>\n" . $this->linefeed;
        }
    }
}

/**
* content for removed data
*/
class zogan_content_removed extends zogan_content {

    function need_indent()
    {
        return true;
    }

    function is_space()
    {
        return true;
    }
}

/**
 * Base class of elements
 * elements are different from other contents in that
 * have child contents.
 * @abstract
 */
class zogan_element extends zogan_content {
    /**
     * child contents in this tag
     * @var array of zogan_content
     */
    var $child = '';

    /**
     * information about variable-data tag
     * 'variable-data tag' is the tag with var="{foo}" attribute.
     * @var array
     */
    var $info;

    /**
     * options
     * 'empty'
     * 'raw'
     * 'allow_delete'
     * @var array
     */
    var $option;

    /**
     * common information in all level of zogan_element
     * 'format':
     * @var array
     */
    var $common;

    /**
     * name of tag name
     * @access private
     */
    var $_tag_name;
    /**
     * attribute in the open tag
     * @var array
     * @access private
     */
    var $_attrs;
    /**
     * parsed attribute value for variable content
     * @var array
     * @access private
     */
    var $_parsed;
    /**
     * if this tag can have child contents or not
     * @var bool
     * @access private
     */
    var $_empty_tag = false;
    /**
     * if content is to be formated or not
     * @access private
     */
    var $_format = null;
    /**
     * @access private
     */
    var $_linefeed_open_tag = '', $_indent_close_tag = '';
    /**
     * @access private
     */
    var $_content = '';

    /**
     * initialize when the tag is opened
     */
    function open($name, $attrs, $parsed, &$info)
    {
        $this->_tag_name = $name;
        $this->_attrs = $attrs;
        $this->_parsed = $parsed;
        $this->child = array();
    }

    /**
     * parse all the data in $this->child stack
     * and set its compiled PHP script into $this->_content.
     * $this->_format
     * @access public
     */
    function close($empty_tag)
    {
        $this->_empty_tag = $empty_tag; // bool

        if (count($this->child) == 0
        || $this->_empty_tag == true) {
            $this->_content = '';
        } else {
            // formatable?
            if (($format = $this->_is_formatable()) !== false) {
                $this->_format = $format;
                $this->_content = '';
                return;
            }

            $num = count($this->child) -1;

            // for pritty print, make indent
            $this->_indent_close_tag = '';
            if ($this->need_indent()) {
                $indent = $this->_get_indent($num+1);
                if ($indent !== false) {
                    foreach ($indent as $j) {
                        $this->_indent_close_tag .= $this->child[$j]->build();
                        $this->child[$j] = new zogan_content_null;
                    }
                }
            }
            for ($i = $num; $i >= 0; $i--) {
                if ($this->child[$i]->need_indent()) {
                    if (($indent = $this->_get_indent($i)) !== false
                    && ($lf = $this->_get_linefeed($i)) !== false) {
                        foreach ($indent as $j) {
                            $this->child[$i]->indent .= $this->child[$j]->build();
                            $this->child[$j] = new zogan_content_null;
                        }
                        foreach ($lf as $j) {
                            $this->child[$i]->linefeed .= $this->child[$j]->build();
                            $this->child[$j] = new zogan_content_null;
                        }
                    }
                }
            }
            $this->_linefeed_open_tag = '';
            $lf = $this->_get_linefeed(-1);
            if ($lf !== false) {
                foreach ($lf as $j) {
                    $this->_linefeed_open_tag .= $this->child[$j]->build();
                    $this->child[$j] = new zogan_content_null;
                }
            }

            // parsing in consideration of altarnative feature
            $this->_content = '';
            for ($i = 0; $i <= $num; $i++) {
                if ($this->child[$i]->is_alternative() !== false) {
                    $i = $this->_parse_alternative($i);
                } else {
                    $this->_content .= $this->child[$i]->build();
                }
            }
        }
    }

    /**
     * _is_formatable()
     * @access private
     * @return array|false array for format
     */
    function _is_formatable()
    {
        $num = count($this->child);
        $format = false;
        for($i = 0; $i < $num; $i++) {
            if ($format == false
            && ($tmp = $this->child[$i]->formatable()) != false) {
                $format = $tmp;
                continue;
            }
            elseif ($this->child[$i]->is_space() != false) {
                continue;
            }
            else {
                return false;
            }
        }
        return $format;
    }

    /**
     * get white spaces for indent
     * @access private
     * @return array index of contents for indent
     */
    function _get_indent($i)
    {
        $indent = array();
        for ($j = $i - 1; $j >= 0; $j--){
            if (is_a($this->child[$j], 'zogan_content_linefeed')) {
                break;
            }
            elseif ($this->child[$j]->is_space()) {
                $indent[] = $j;
                continue;
            }
            else {
                return false;
            }
        }
        if ($j < 0) {
            return false;
        }
        return $indent;
    }

    /**
     * get white spaces and linefeed after tag for pretty print
     * @access private
     * @return array index of contents
     */
    function _get_linefeed($i)
    {
        $num = count($this->child) -1;
        $lf = array();
        for ($j = $i + 1; $j <= $num; $j++){
            if (is_a($this->child[$j], 'zogan_content_linefeed')) {
                $lf[] = $j;
                break;
            }
            elseif ($this->child[$j]->is_space()) {
                $lf[] = $j;
                continue;
            }
            else {
                return false;;
            }
        }
        if ($j > $num) {
            return false;
        }
        return $lf;
    }

    /**
     * parse the tag with alternative
     * @access private
     * @return string
     */
    function _parse_alternative($i)
    {
        $num = count($this->child);
        $alts = array($i);
        $sign = $this->child[$i]->is_alternative();
        for ($i++; $i < $num; $i++){
            if ($this->child[$i]->is_alternative() === $sign) {
                $alts[] = $i;
                continue;
            }
            elseif ($this->child[$i]->is_space() == true) {
                continue;
            }
            break;
        }

        $level      = $this->child[$alts[0]]->info['level'];
        $_var       = $this->child[$alts[0]]->info['name'];
        $_each      = '$_each' . $level;
        $_valprev   = '$_val'  . ($level -1);
        $_valcurr   = '$_val'  . $level;
        $_num       = '$_num'  . $level;
        $_alts      = count($alts);

        $built
            = '<?php '
            . "$_each = \$this->_c_each($_valprev, '$_var'); "
            . "$_num = 0; "
            . "foreach ($_each as $_valcurr) { ";

        $_j = 0;
        foreach ($alts as $i) {
            $built
                .= "if ($_num%$_alts == $_j) { "
                . $this->child[$i]->parse_in_loop()
                . "continue; } ";
            $_j++;
        }
        $built .= "} ?>\n";

        $this->_content .= $built;

        return end($alts);
    }

    /**
     * returns build tag
     * @return array array($open_tag, $close_tag)
     * @access private
     */
    function _build_tag()
    {
        // omit?
        if (@$this->_parsed[ZOGAN_OMIT_TAG] === true) {
            $open_tag = '';
            $close_tag = '';
            return;
        }

        // open tag
        $open_tag = '<' . $this->_tag_name;
        $_val = '$_val' . ($this->info['level'] -1);

        foreach ($this->_attrs as $key => $value) {

            if (is_array($value)) {
            // variable attr
                $_key = $value['key'];
                $_varibale = $value['var'];
                $_default = $this->_single_quote($value['default']);

                $get_val = "\$this->_c_val($_val, '$_varibale')";
                if ($value['func'] == '!') {
                    $get_val = '!(bool)'.$get_val;
                }
                elseif ($value['func'] == '?') {
                    $get_val = '(bool)'.$get_val;
                }
                elseif ($value['func'] != '') {
                    $_func = $value['func'];
                    $get_val = "\$this->_c_func($get_val, '$_func', true)";
                }

                $open_tag .= "<?php echo \$this->_c_attr('$_key', $get_val, $_default); ?>\n";
            }
            elseif (is_string($value)) {
            // static attr
                $open_tag .= ' ' . $key . '=' . $this->_double_quote_xml($value);
            }
            elseif ($value === true) {
            // var:cont attr
                $_val = '$_val' . $this->info['level'];
            }
        }

        if ($this->_empty_tag == true) {
            $open_tag  .= ' />';
            $close_tag  = '';
        } else {
            $open_tag  .= '>';
            $close_tag  = '</' . $this->_tag_name . '>';
        }

        return array($open_tag, $close_tag);
    }

    /**
     * returns namespace information for ZOGAN
     * @return array
     */
    function getNamespace()
    {
        return $this->_parsed[ZOGAN_NAMESPACE];
    }

    /**
     * returns name of tag
     * @return array
     */
    function getTagname()
    {
        return $this->_tag_name;
    }

    /**
     * build when ZOGAN_OMIT=='first'
     * @return string
     */
    function _omit_first($built)
    {
        if (@$this->_parsed[ZOGAN_OMIT] == 'first') {
            $_num   = '$_num'  . $this->info['level'];
            return "<?php if ($_num>1) { ?>\n". $built . "<?php } ?>\n";
        } else {
            return $built;
        }
    }
}

/**
 * Parser for root level of the stack
 *
 * @package zogan
 */
class zogan_element_root extends zogan_element {

    function zogan_element_root()
    {
        $this->child = array();
        $this->_parsed[ZOGAN_NAMESPACE]
            = array(ZOGAN_PREFIX => ZOGAN_NAMESPACE);
        $this->info
            = array('level' => 0, 'nests' => false);
    }

    function close($close_tag)
    {
        // do nothing. In fact, this method is not called.
        return;
    }

    function build()
    {
        parent::close(false);
        foreach ($this->common['format'] as $type) {
            $this->_content =
                "<?php require_once ZOGAN_FORMAT.'$type'.'.php'; ?>\n"
                . $this->_content;
        }
        return $this->_content;
    }
}


/**
 * Parser for tags without variable attribute
 *
 * @package zogan
 */
class zogan_element_static extends zogan_element {

    function open($name, $attrs, $parsed, &$info)
    {
        $this->info =& $info;
        parent::open($name, $attrs, $parsed, $info);
    }

    function build()
    {

        list($open_tag, $close_tag) = $this->_build_tag();
        $open_tag = $this->indent . $open_tag;
        $close_tag = $close_tag . $this->linefeed;
        $built = $open_tag . $this->_linefeed_open_tag
                . $this->_content
                . $this->_indent_close_tag . $close_tag;
        $built = $this->_omit_first($built);
        return $built;
    }

    function need_indent()
    {
        return isset($this->_parsed[ZOGAN_OMIT]);
    }
}

/**
 * Parser for variable tag
 * variable tag means the one with var attribute.
 *
 * @package zogan
 */
class zogan_element_variable extends zogan_element {

    function open($name, $attrs, $parsed, &$info)
    {
        $this->info['level'] = (int)$info['level'] +1;
        $this->info['nests'] = false;
        $info['nests'] = true;
        $this->info['name'] = $parsed[ZOGAN_CONTENT]['var'];
        $this->info['func'] = $parsed[ZOGAN_CONTENT]['func'];
        parent::open($name, $attrs, $parsed, $info);
    }

    function build()
    {
        $_var       = $this->info['name'];
        $_each      = '$_each' . $this->info['level'];
        $_valprev   = '$_val'  . ($this->info['level'] -1);
        $_valcurr   = '$_val'  . $this->info['level'];
        $_num       = '$_num'  . $this->info['level'];

        $built
            = '<?php '
            . "$_each = \$this->_c_each($_valprev, '$_var'); "
            . "$_num = 0; "
            . "foreach ($_each as $_valcurr) { "
            . $this->parse_in_loop()
            . "} ?>\n";

        return $built;
    }

    function parse_in_loop()
    {
        list($open_tag, $close_tag) = $this->_build_tag();
        $open_tag = $this->indent . $open_tag;
        $close_tag = $close_tag . $this->linefeed;

        $_var       = $this->info['name'];
        $_valcurr   = '$_val'  . $this->info['level'];
        $_num       = '$_num'  . $this->info['level'];
        $escape     = true;
        $built      = '';

        if ($this->info['func'] == '!') {
            $built .= "$_valcurr = !(bool)$_valcurr; ";
        }
        elseif ($this->info['func'] == '?') {
            $built .= "$_valcurr = (bool)$_valcurr; ";
        }
        elseif ($this->info['func'] != '') {
            $_func = $this->info['func'];
            $built .= "$_valcurr = \$this->_c_func($_valcurr, '$_func'); ";
            $escape = false;
        }

        $built
            .= "if ($_valcurr === null || $_valcurr === false) { "
            . '$this->_c_nodata('
            . $this->_double_quote($this->indent).', '
            . $this->_double_quote($this->linefeed).'); '
            . '} else { '
            . "$_num++; ";

        if (@$this->_parsed[ZOGAN_OMIT] == 'first') {
            $built .= "if ($_num>1) { ";
        }

        if (! $this->info['nests']) {
        // not nested
            if (! isset($this->_format)) {
            // normal data
                $built
                   .= "?>\n"
                    . $open_tag
                    . "<?php if (\$this->_c_is_string($_valcurr)) { ";
                $built
                    .= ($escape)
                        ? "echo \$this->_c_escape($_valcurr, '$_var'); } "
                        : "echo $_valcurr; } ";
                $built
                    .= "elseif ($_valcurr === true) { ?>\n"
                    . $this->_linefeed_open_tag . $this->_content . $this->_indent_close_tag
                    . "<?php } else { "
                    . "\$this->_c_warn(); "
                    . "} ?>\n"
                    . $close_tag;
            } else {
            // data to be formated
                $_func = $this->_format[0];
                $_format = $this->_single_quote($this->_format[1]);
                $this->common['format'][$this->_format[0]] = $this->_format[0];
                $built
                   .= "?>\n"
                    . $open_tag
                    . "<?php if (! zogan_format_$_func($_valcurr, $_format)) { "
                    . "\$this->_c_warn(); } ?>\n"
                    . $close_tag;
            }
        } else {
        // nested
            $built
                .= "\$this->_c_hash($_valcurr, '$_var'); ?>\n"
                 . $open_tag . $this->_linefeed_open_tag
                 . $this->_content
                 . $this->_indent_close_tag . $close_tag;
        }

        $built
            .= '<?php ';

        if (@$this->_parsed[ZOGAN_OMIT] == 'first') {
            $built .= '} ';
        }

        $built
            .= '} ';

        return $built;
    }

    function need_indent()
    {
        return true;
    }

    function is_alternative()
    {
        if (@$this->_parsed[ZOGAN_ALTERNATON] === true) {
            return array($this->_tag_name, $this->info['name']);
        } else {
            return false;
        }
    }
}

/**
 * Parser for tags with '#num' of attribute for variable
 * which shows the number of iteration.
 * @package zogan
 */
class zogan_element_num extends zogan_element_static {

    function build()
    {
        list($open_tag, $close_tag) = $this->_build_tag();
        $open_tag = $this->indent . $open_tag;
        $close_tag = $close_tag . $this->linefeed;

        $_num = '$_num' . $this->info['level'];
        $built = $open_tag . $this->_linefeed_open_tag
                . "<?php echo $_num; ?>\n"
                . $this->_indent_close_tag . $close_tag;
        $built = $this->_omit_first($built);
        return $built;
    }
}

/**
 * include other template
 * @package zogan
 * @todo when non existant file?
 */
class zogan_element_inc extends zogan_element_static {

    function build()
    {
        list($open_tag, $close_tag) = $this->_build_tag();
        $open_tag = $this->indent . $open_tag;
        $close_tag = $close_tag . $this->linefeed;

        $_valcurr   = '$_val'  . $this->info['level'];
        $_zogan    = '$_ppt'  . $this->info['level'];
        $_incfile = $this->_parsed[ZOGAN_CONTENT]['arg'];

        $built = $open_tag . $this->_linefeed_open_tag
           . "<?php "
           . "$_zogan = zogan_clone_obj(\$this); "
           . $_zogan."->_c_change_basename(".$this->_single_quote($_incfile)."); "
           . $_zogan."->display($_valcurr); "
           . "unset($_zogan); ?>\n"
           . $this->_indent_close_tag. $close_tag;

        $built = $this->_omit_first($built);
        return $built;
    }
}

/**
 * omitted element
 */
class zogan_element_omit extends zogan_element_static {
    function build()
    {
        return '';
    }

    function need_indent()
    {
        return true;
    }

    function is_space()
    {
        return true;
    }
}

?>